/**************************************************************************
*
*	Main.cpp
*
*	Main source file for simple robot sample programme
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

/*************************************************************************/
// Include files
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>

#include "Controller.h"
#include "Graphics.h"

/*************************************************************************/
// Defines
/*************************************************************************/

// IDs to use for debug graphics
#define GRAPHIC_ID_PLAYING_FIELD		1
#define GRAPHIC_ID_ROBOT				2
#define GRAPHIC_ID_SEARCH_TARGET		3
#define GRAPHIC_ID_ENEMY				4

/*************************************************************************/
// Local types
/*************************************************************************/

/*************************************************************************/
// Local functions
/*************************************************************************/

// RandomBetween() - Helper function to return a random value between the 
// two bounds.
static double RandomBetween(double LowerBound, double UpperBound);

/*************************************************************************/
// Functions
/*************************************************************************/

/**************************************************************************
*
*	function:	main()
*
*	desc:		Programme entry point
*
*	notes:		
*
**************************************************************************/

void main(int NumArgs, char** Args)
{
	try
	{
		printf("Simple Robot Sample\n\n");

		// seed random number generator
		srand((unsigned int)time(NULL));

		// setup default values of parameters
		std::string ServerName = "localhost";
		bool DebugGraphics = false;

		// expect arguments to be in the following format:
		// /a a_value /b b_value
		// so there should be an even number of them, after the first one
		// which is the exe filename
		if ((NumArgs - 1) % 2 != 0)
			throw std::runtime_error("Expected an even number of arguments");
		
		// now loop over them
		int i;
		for (i = 1; i < NumArgs; i += 2)
		{
			if (!strcmp(Args[i], "/s"))
			{
				// /s = server address (a string)
				ServerName = Args[i + 1];
			}
			else if (!strcmp(Args[i], "/d"))
			{
				// /d = use debug graphics (an int, 1 = yes, 0 = no)
				DebugGraphics = (atoi(Args[i + 1]) != 0);
			}
			else
			{
				throw std::runtime_error(
					std::string() + "Invalid argument " + Args[i]);
			}
		}

		// create a controller
		Controller_t Controller;

		// connect to the server
		Controller.Connect(ServerName, 9386);

		// create a graphics client
		Graphics_t Graphics;

		// if debugging, connect (on localhost)
		if (DebugGraphics)
			Graphics.Connect("localhost", 9387);

		// get the playing field size
		double PlayingFieldWidth, PlayingFieldHeight;
		Controller.GetPlayingFieldSize(
			PlayingFieldWidth, PlayingFieldHeight);
		printf("Playing field size is %.1fm x %.1fm\n", 
			PlayingFieldWidth, PlayingFieldHeight);

		// if debugging, add the playing field for reference
		if (DebugGraphics)
		{
			Graphics.SetToRectangle(
				GRAPHIC_ID_PLAYING_FIELD,
				Vec_t(0, 0),				// centre position
				PlayingFieldWidth,
				PlayingFieldHeight,
				0,							// orientation
				4,							// line width
				0, 0, 139);					// colour
		}

		// create a robot
		int RobotID = Controller.CreateRobot();
		printf("Robot ID is %d\n", RobotID);

		// set name and owner
		Controller.SetName("Simple Robot");
		Controller.SetOwner("Saxon");

		// set colour
		Controller.SetColour(0, 0, 255);

		// get initial amount of money
		printf("Robot has $%d to spend on equipment\n", 
			Controller.GetMoney());

		// buy equipment
		Controller.PurchaseEngine(3, 75);
		Controller.PurchaseSensor(10, 60, 60);
		Controller.PurchaseBattery(50);
		Controller.PurchaseTurret(0, 0);
		Controller.PurchaseWeapon(1.3, 4, 2, 4.5);
		Controller.PurchaseSolarPanel(1.5);

		// get amount of money left
		printf("Robot has $%d remaining after purchasing equipment\n", 
			Controller.GetMoney());

		// enter the simulation
		Controller.EnterSimulation();

		// set up a value to hold the position which the robot is
		// trying to get to, and initialise it with the robot's
		// current position
		Vec_t TargetPosition = Controller.GetPosition();

		// begin assuming that we haven't killed anybody yet
		int LastNumKills = 0;

		try
		{
			// now loop until a key is pressed
			printf("\nHit a key to stop\n\n");
			while (!_kbhit())
			{
				// get the robot's current position
				Vec_t CurrentPosition = Controller.GetPosition();

				// get the robot's current orientation
				double CurrentOrientation = Controller.GetOrientation();

				// if debugging, draw the robot
				if (DebugGraphics)
				{
					Graphics.SetToRectangle(
						GRAPHIC_ID_ROBOT,
						CurrentPosition,
						1.0,					// length
						0.6,					// 'height' (width of robot)
						CurrentOrientation,
						1,						// line width
						0, 0, 255);				// colour
				}

				// sense other robots
				std::vector<Controller_t::SensedRobot_t> Robots = 
					Controller.GetSensedRobots();

				if (Robots.empty())
				{
					// we couldn't see any robots, so simply wander randomly
					// around the playing field, in the hope that we'll
					// eventually find another robot

					// work out the vector from the current position to the
					// target position
					Vec_t TargetVector = TargetPosition - CurrentPosition;

					if (TargetVector.Length() < 1.0)
					{
						// we're within 1 metre of the target, so assume we've
						// reached it - therefore pick a new random target
						// position, which is at least 2 metres from the edge
						// of the playing field
						TargetPosition.x = RandomBetween(
							-(PlayingFieldWidth / 2.0 - 2.0), 
							(PlayingFieldWidth / 2.0 - 2.0));
						TargetPosition.y = RandomBetween(
							-(PlayingFieldHeight / 2.0 - 2.0), 
							(PlayingFieldHeight / 2.0 - 2.0));

						printf(
							"New target position is %.1f, %.1f\n",
							TargetPosition.x,
							TargetPosition.y);
						
						// recalculate the target vector since it is used below
						TargetVector = TargetPosition - CurrentPosition;
					}

					// work out the desired orientation to point towards the
					// target
					double DesiredOrientation = TargetVector.GetAngle();

					// set it
					Controller.SetDesiredOrientation(
						DesiredOrientation);

					// work out the difference between the current orientation and
					// the desired orientation
					double OrientationDifference = 
						DesiredOrientation - CurrentOrientation;
					
					// wrap to -pi to pi range
					if (OrientationDifference < -M_PI)
						OrientationDifference += 2 * M_PI;
					else if (OrientationDifference >= M_PI)
						OrientationDifference -= 2 * M_PI;

					if (fabs(OrientationDifference) < 5 * M_PI / 180)
					{
						// we're within 5 degrees of the desired orientation, so
						// set the throttle to full forwards
						Controller.SetThrottle(100.0);
					}
					else
					{
						// set the throttle to stay still while we turn
						Controller.SetThrottle(0.0);
					}

					// set sensor orientation to 90 degrees further clockwise
					// than it currently is, which should result in a constant
					// clockwise spin as fast as possible (assuming we aren't
					// able to spin by 90 degrees between iterations of this loop)
					Controller.SetDesiredSensorOrientation(
						Controller.GetSensorOrientation() + M_PI / 2);

					// if debugging, draw a circle at the point we're targeting,
					// and hide any indication of an enemy
					if (DebugGraphics)
					{
						Graphics.SetToCircle(
							GRAPHIC_ID_SEARCH_TARGET,
							TargetPosition,
							1.0,				// radius
							1,					// line width
							50, 200, 50);		// colour

						Graphics.Remove(GRAPHIC_ID_ENEMY);
					}
				}
				else
				{
					// we can see at least one robot, so attack the first one

					// work out the vector from the current position to the
					// enemy robot position
					Vec_t EnemyVector = Robots[0].m_Position - CurrentPosition;

					// work out the desired orientation to point towards the
					// enemy robot
					double DesiredOrientation = EnemyVector.GetAngle();

					// set it
					Controller.SetDesiredOrientation(
						DesiredOrientation);

					// work out the difference between the current orientation and
					// the desired orientation
					double OrientationDifference = 
						DesiredOrientation - CurrentOrientation;
					
					// wrap to -pi to pi range
					if (OrientationDifference < -M_PI)
						OrientationDifference += 2 * M_PI;
					else if (OrientationDifference >= M_PI)
						OrientationDifference -= 2 * M_PI;

					if (fabs(OrientationDifference) < 5 * M_PI / 180)
					{
						// we're within 5 degrees of the desired orientation

						// work out how far we are from the enemy
						double EnemyDistance = EnemyVector.Length();

						if (EnemyDistance < 5)
						{
							// within 5 metres of the enemy, so move
							// backwards to keep away from it (we'll
							// be too easy to hit if we get too close)
							Controller.SetThrottle(-100.0);
						}
						else if (EnemyDistance < 6)
						{
							// between 5 and 6 metres from the enemy,
							// so stay still
							Controller.SetThrottle(0.0);
						}
						else
						{
							// more than 6 metres from the enemy, so
							// move towards it
							Controller.SetThrottle(100.0);
						}

						if (EnemyDistance < 6)
						{
							// we're within a reasonable range of the enemy,
							// so fire the weapon if it's fully charged
							if (Controller.GetWeaponCharge() >= 100)
								Controller.FireWeapon();
						}
					}
					else
					{
						// set the throttle to stay still while we turn
						Controller.SetThrottle(0.0);
					}

					// set the sensor angle required to keep it pointing 
					// at the enemy, to make sure it doesn't get away
					Controller.SetDesiredSensorOrientation(
						DesiredOrientation - CurrentOrientation);

					// if debugging, draw a circle at the point of the enemy,
					// and hide any search target
					if (DebugGraphics)
					{
						Graphics.SetToCircle(
							GRAPHIC_ID_ENEMY,
							Robots[0].m_Position,
							0.5,				// radius
							1,					// line width
							255, 0, 0);			// colour

						Graphics.Remove(GRAPHIC_ID_SEARCH_TARGET);
					}
				}

				// check if we've killed anybody else
				int NumKills = Controller.GetNumKills();
				if (LastNumKills != NumKills)
				{
					Controller.Taunt("Woohoo I killed somebody!");
					LastNumKills = NumKills;
				}
			}
			// read the pressed key
			_getch();
		}
		catch (...)
		{
			// check if the exception was because we tried to
			// control the robot after it has been killed
			if (Controller.GetHealth() <= 0)
			{
				// we've been killed - get details of our killer
				int KillerID;
				std::string KillerName, KillerOwner;
				Controller.GetKiller(KillerID, KillerName, KillerOwner);

				// then send a parting message
				Controller.Taunt(
					"I don't like you " + 
					KillerName + " / " + KillerOwner + " !!");
			}
			else
			{
				// the exception was for some other reason, so propagate it
				throw;
			}
		}

		printf("\nDone\n\n");
	}
	catch (std::exception& e)
	{
		printf("\nAn exception occurred: %s\n\n", e.what());
	}
}

/**************************************************************************
*
*	function:	RandomBetween()
*
*	desc:		Helper function to return a random value between the two 
*				bounds.
*
*	notes:		
*
**************************************************************************/

static double RandomBetween(double LowerBound, double UpperBound)
{
	return LowerBound + rand() * (UpperBound - LowerBound) / RAND_MAX;
}

/*************************************************************************/
